@@ -216,9 +216,15 @@ p11_mmap_open (const char *path,
 		return NULL;
 	}
 
+	if (sb->st_size == 0) {
+		*data = "";
+		*size = 0;
+		return map;
+	}
+
 	map->size = sb->st_size;
 	map->data = mmap (NULL, map->size, PROT_READ, MAP_PRIVATE, map->fd, 0);
-	if (map->data == NULL) {
+	if (map->data == MAP_FAILED) {
 		close (map->fd);
 		free (map);
 		return NULL;
@@ -232,7 +238,8 @@ p11_mmap_open (const char *path,
 void
 p11_mmap_close (p11_mmap *map)
 {
-	munmap (map->data, map->size);
+	if (map->size)
+		munmap (map->data, map->size);
 	close (map->fd);
 	free (map);
 }
@@ -512,28 +519,11 @@ strconcat (const char *first,
 
 #endif /* HAVE_STRCONCAT */
 
-#ifndef HAVE_ASPRINTF
-
-int
-asprintf (char **strp,
-          const char *fmt,
-          ...)
-{
-	va_list va;
-	int ret;
-
-	va_start (va, fmt);
-	ret = vasprintf (strp, fmt, va);
-	va_end (va);
-
-	return ret;
-}
-
-#endif /* HAVE_ASPRINTF */
-
 #ifndef HAVE_VASPRINTF
 #include <stdio.h>
 
+int vasprintf(char **strp, const char *fmt, va_list ap);
+
 int
 vasprintf (char **strp,
            const char *fmt,
@@ -578,6 +568,27 @@ vasprintf (char **strp,
 
 #endif /* HAVE_VASPRINTF */
 
+#ifndef HAVE_ASPRINTF
+
+int asprintf(char **strp, const char *fmt, ...);
+
+int
+asprintf (char **strp,
+          const char *fmt,
+          ...)
+{
+	va_list va;
+	int ret;
+
+	va_start (va, fmt);
+	ret = vasprintf (strp, fmt, va);
+	va_end (va);
+
+	return ret;
+}
+
+#endif /* HAVE_ASPRINTF */
+
 #ifndef HAVE_GMTIME_R
 
 struct tm *
@@ -833,7 +844,22 @@ strerror_r (int errnum,
             size_t buflen)
 {
 #ifdef OS_WIN32
+#if _WIN32_WINNT < 0x502 /* WinXP or older */
+	int n = sys_nerr;
+	const char *p;
+	if (errnum < 0 || errnum >= n)
+		p = sys_errlist[n];
+	else
+		p = sys_errlist[errnum];
+	if (buf == NULL || buflen == 0)
+		return EINVAL;
+	strncpy(buf, p, buflen);
+	buf[buflen-1] = 0;
+	return 0;
+#else /* Server 2003 or newer */
 	return strerror_s (buf, buflen, errnum);
+#endif /*_WIN32_WINNT*/
+
 #else
 	#error no strerror_r implementation
 #endif
